/*
                  Search and Rescue Compile Time Options
 */


#ifndef CONFIG_H
#define CONFIG_H

/*
 *      Default size (in pixels) of the toplevel game iwndow:
 */
#define SAR_DEF_WINDOW_WIDTH    320
#define SAR_DEF_WINDOW_HEIGHT   240


/*
 *	Maximum length of an object name in bytes (include null terminating
 *	byte):
 */
#define SAR_OBJ_NAME_MAX	256

/*
 *	Maximum literal game command (not including argument) length:
 */
#define SAR_CMD_MAX		80

/*
 *      Default directory paths:
 */

/* Local game dir, user's $HOME directory will be prefixed to this. */
#define SAR_DEF_LOCAL_DATA_DIR  ".SearchAndRescue"

/* Global game dir. */
#define SAR_DEF_GLOBAL_DATA_DIR "/usr/share/games/SearchAndRescue"

/* Subdirs, one of each in the local and global game dirs. */
#define SAR_DEF_AIRCRAFTS_DIR		"aircrafts"
#define SAR_DEF_CONTROL_PANELS_DIR	"control_panels"
#define SAR_DEF_IMAGES_DIR		"images"
#define SAR_DEF_MISSIONS_DIR		"missions"
#define SAR_DEF_OBJECTS_DIR		"objects"
#define SAR_DEF_SCENERY_DIR		"scenery"
#define SAR_DEF_SOUNDS_DIR		"sounds"
#define SAR_DEF_TEXTURES_DIR		"textures"


/*
 *	Default file paths:
 */

/* Run time preferances file. */
#define SAR_DEF_RC_FILE			"SearchAndRescue.ini"

/* Texture referance names and file names list file. */
#define SAR_DEF_TEXTURES_FILE		"textures.ini"

/* Weather definations file. */
#define SAR_DEF_WEATHER_FILE		"weather.ini"

/* Icon paths. */
#ifdef __MSW__
# define SAR_DEF_SAR_ICON_FILE		"SearchAndRescue.ico"
#else
# define SAR_DEF_SAR_ICON_FILE		"/usr/share/icons/SearchAndRescue.xpm"
#endif

/* Standard texture referance names that must be defined in
 * the SAR_DEF_TEXTURES_FILE.
 */
#define SAR_STD_TEXNAME_BASKET_BOTTOM	"sar_basket_bottom"
#define SAR_STD_TEXNAME_BASKET_END	"sar_basket_end"
#define SAR_STD_TEXNAME_BASKET_SIDE	"sar_basket_side"
#define SAR_STD_TEXNAME_EXPLOSION	"sar_explosion"
#define SAR_STD_TEXNAME_HELIPAD		"sar_helipad"
#define SAR_STD_TEXNAME_RUNWAY		"sar_runway"
#define SAR_STD_TEXNAME_ROTOR_WASH	"sar_rotor_wash"
#define SAR_STD_TEXNAME_SMOKE_LIGHT	"sar_smoke_light"
#define SAR_STD_TEXNAME_SMOKE_DARK	"sar_smoke_dark"
#define SAR_STD_TEXNAME_WATER_RIPPLE	"sar_water_ripple"

/* Menu background images. */
#define SAR_DEF_MENU_BGIMG_STANDARD_FILE	"images/menu_std_bg.tga"
#define SAR_DEF_MENU_BGIMG_FREE_FLIGHT_FILE	"images/menu_free_flight.tga"
#define SAR_DEF_MENU_BGIMG_MAIN_FILE		"images/menu_main.tga"
#define SAR_DEF_MENU_BGIMG_MISSION_FILE		"images/menu_mission.tga"
#define SAR_DEF_MENU_BGIMG_OPTIONS_FILE		"images/menu_options.tga"
#define SAR_DEF_MENU_BGIMG_PROGRESS_FILE	"images/menu_progress.tga"

/* Mission map images. */
#define SAR_DEF_MISSION_MAP_IMG_HELIPAD_FILE	"images/mm_helipad.tga"
#define SAR_DEF_MISSION_MAP_IMG_INTERCEPT_FILE	"images/mm_intercept.tga"
#define SAR_DEF_MISSION_MAP_IMG_HELICOPTER_FILE	"images/mm_helicopter.tga"
#define SAR_DEF_MISSION_MAP_IMG_VICTIM_FILE	"images/mm_victim.tga"
#define SAR_DEF_MISSION_MAP_IMG_BOAT_FILE	"images/mm_boat.tga"


/* Sound objects. */
#define SAR_DEF_SOUND_LAND_WHEEL_SKID	"sounds/land_wheel_skid.wav"
#define SAR_DEF_SOUND_LAND_SKI_SKID	"sounds/land_ski_skid.wav"
#define SAR_DEF_SOUND_LAND_SKI		"sounds/land_ski.wav"
#define SAR_DEF_SOUND_LAND_BELLY	"sounds/land_belly.wav"
#define SAR_DEF_SOUND_CRASH_OBSTRUCTION	"sounds/crash_obstruction.wav"
#define SAR_DEF_SOUND_CRASH_GROUND	"sounds/crash_ground.wav"

/*
 *	Comment character (for program's file formats):
 */
#define SAR_COMMENT_CHAR	'#'

/*
 *	Program configuration deliminator character.
 */
#define SAR_CFG_DELIM_CHAR	'='


/*
 *      Show messages for this many milliseconds when a new one is
 *      printed.
 */
#define SAR_MESSAGE_SHOW_INT    5000


/*
 *	Human contact cylendrical bounds (origin at base of feet)
 *	in meters:
 */
#define SAR_HUMAN_CONTACT_RADIUS	1.2
#define SAR_HUMAN_CONTACT_ZMAX		2.5
#define SAR_HUMAN_CONTACT_ZMIN		0.0


/*
 *      Smoke puff respawn interval in milliseconds:
 */
#define DEF_SMOKE_SPAWN_INT     3000

/*
 *      Smoke puff life span in milliseconds:
 */
#define DEF_SMOKE_LIFE_SPAN     12000

/*
 *      Explosion frame increment interval in milliseconds:
 */
#define DEF_EXPLOSION_FRAME_INT 500

/*
 *      Explosions caused by crashes last this long in milliseconds:
 */
#define DEF_CRASH_EXPLOSION_LIFE_SPAN   60000

/*
 *	Rotor wash visiblity coefficient.
 */
#define DEF_ROTOR_WASH_VIS_COEFF	0.5

/*
 *	Default ground base graphics tiling (in meters):
 */
#define DEF_GROUND_BASE_TILE_WIDTH		400
#define DEF_GROUND_BASE_TILE_HEIGHT		400
#define DEF_GROUND_BASE_TILE_CLOSE_RANGE	800


/*
 *	Default cloud layer graphics tiling (in meters):
 */
#define DEF_CLOUD_TILE_WIDTH		8000
#define DEF_CLOUD_TILE_HEIGHT		8000
#define DEF_CLOUD_TILE_CLOSE_RANGE	16000


/*
 *	Default planetary measurements:
 */
#define SAR_DEF_PLANET_RADIUS		6367470.96322	/* Meters. */

/*
 *	Human animation rate (in animation units per cycle):
 */
#define SAR_HUMAN_ANIM_RATE	50000


/*
 *	Keyboard functions help message (array of lines):
 */
#define SAR_KEYS_HELP_MESSAGES	{ \
 "F1  Display key functions.", \
 "F2  Cockpit view.", \
 "F3  Spot view.", \
 "F4  Fly by tower view.", \
 "F5  Rescue basket view.", \
 "CURSORS  Attitude control.", \
 "SHIFT + CURSORS  Move/orbit view.", \
 "BKSP  Normalize view.", \
 "PGUP/PGDN  Throttle.", \
 "SHIFT + PGUP/PGDN  Throttle fast.", \
 "CTRL + CURSORS  Rudder/steering.", \
 "-/+  Zoom in/out.", \
 "CTRL + -/+  Hoist in/out.", \
 "g  Raise/lower landing gears.", \
 "d  Open/close door.", \
 "o  Strobes on/off.", \
 "l  Lights on/off.", \
 "h  HUD and text brighten.", \
 "SHIFT + h HUD and text darken.", \
 "m  Map.", \
 "w  Select waypoint.", \
 "s  Scores/mission status.", \
 ".  Wheel brakes.", \
 "SHIFT + . Parking brakes.", \
 "z  Decelerate time (slow-motion).", \
 "Z  Accelerate time (compress).", \
 "/  Literal OOC command.", \
 "CTRL + d  Flight physics.", \
 "CTRL + s  Sound level.", \
 "F9  Toggle ground texture.", \
 "F10  Toggle atmosphere.", \
 "F11  Toggle object texture.", \
 "F12  Toggle clouds." \
}

/*
 *	Longest line in SAR_KEYS_HELP_MESSAGES plus a 2 chars tolorance:
 */
#define SAR_KEYS_HELP_MESSAGE_LINE_MAX	35


#endif	/* CONFIG_H */
