#include <stdio.h>
#include <string.h>

#include "v3dtex.h"

#include "obj.h"
#include "explosion.h"
#include "sar.h"


int ExplosionCreatePosition(
	sar_scene_struct *scene,
	sar_object_struct ***ptr, int *total,
        sar_position_struct *pos,
        double radius,
        int ref_object,
        const char *tex_name
);


/*
 *	Creates an explision at the specified position referenced by
 *	ref_object.
 */
int ExplosionCreatePosition(
	sar_scene_struct *scene,
        sar_object_struct ***ptr, int *total,
        sar_position_struct *pos,
        double radius,
	int ref_object,
	const char *tex_name
)
{
	int obj_num;
	sar_object_struct *obj_ptr;
	sar_object_explosion_struct *obj_explosion_ptr;


	if((scene == NULL) ||
           (pos == NULL)
	)
	    return(-1);

	/* Create explosion object. */
	obj_num = SARObjCreate(
	    scene, ptr, total,
	    SAR_OBJ_TYPE_EXPLOSION
	);
	if(obj_num < 0)
	    return(-1);

	obj_ptr = (*ptr)[obj_num];
	if(obj_ptr == NULL)
	    return(-1);


	/* Set new values. */
	memcpy(
	    &obj_ptr->pos,
	    pos,
	    sizeof(sar_position_struct)
	);
	obj_ptr->range = 7000;		/* Visable about 5 miles. */
/*	obj_ptr->life_span = 0;		Calculated farter below. */


	/* Get pointer to explosion sub structure. */
	obj_explosion_ptr = (sar_object_explosion_struct *)
	    obj_ptr->data;
	if(obj_explosion_ptr == NULL)
	    return(-1);

	obj_explosion_ptr->radius = radius;
	obj_explosion_ptr->cur_frame = 0;
	obj_explosion_ptr->next_frame_inc = cur_millitime +
	    option.explosion_frame_int;

	obj_explosion_ptr->tex_num = SARGetTextureRefNumberByName(
	    scene, tex_name
	);
	obj_explosion_ptr->ref_object = ref_object;


	/* Calculate life span. */
	if(obj_explosion_ptr->tex_num > -1)
	{
	    v3d_texture_ref_struct *t;

	    t = scene->texture_ref[obj_explosion_ptr->tex_num];

            obj_ptr->life_span = cur_millitime + (t->total_frames *
		option.explosion_frame_int);
	}

	return(obj_num);
}
