#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>

#include "../include/os.h"

#ifdef __MSW__
# include <windows.h>
#endif
  
#include <GL/gl.h>
         
#include "../include/string.h"  

#include "v3dtex.h"
                
#include "gw.h"
#include "stategl.h"

#include "image.h"
#include "menu.h"


int SARMenuCreateMap(
	sar_menu_struct *m,
	double x, double y,
	double width, double height,
	double scroll_x, double scroll_y,
	double m_to_pixels_coeff
);
void SARMenuMapDeleteAllMarkings(sar_menu_map_struct *map_ptr);
void SARMenuDrawMap(
	gw_display_struct *display,
	sar_menu_struct *m,
	sar_menu_map_struct *map_ptr,
	int map_num
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))


/*
 *	Creates a map object on menu m, returns the index number of
 *	the map object or -1 on failure.
 */
int SARMenuCreateMap(
        sar_menu_struct *m,
        double x, double y,
        double width, double height,
	double scroll_x, double scroll_y,
        double m_to_pixels_coeff
)
{
	int i, n;
        sar_menu_map_struct *map_ptr;


        if(m == NULL)
            return(-1);

        if(m->total_objects < 0)
            m->total_objects = 0;

        for(i = 0; i < m->total_objects; i++)
        {
            if(m->object[i] == NULL)
                break;
        }
        if(i < m->total_objects)
        {
            n = i;
        }
        else
        {
            n = m->total_objects;
            m->total_objects++;

            m->object = (void **)realloc(
                m->object,
                m->total_objects * sizeof(void *)
            );
            if(m->object == NULL)
            {
                m->total_objects = 0;
                return(-1);
            }

            m->object[n] = NULL;
        }   
            
        /* Allocate structure. */
        map_ptr = (sar_menu_map_struct *)calloc(
            1,
            sizeof(sar_menu_map_struct)
        );
        if(map_ptr == NULL)
            return(-1);
        else
            m->object[n] = (void *)map_ptr;
                
         
        /* Set map values. */
        map_ptr->type = SAR_MENU_OBJECT_TYPE_MAP;
        map_ptr->x = x;
        map_ptr->y = y;
        map_ptr->width = width;
        map_ptr->height = height;
	map_ptr->scroll_x = (int)(-scroll_x * m_to_pixels_coeff);
        map_ptr->scroll_y = (int)(-scroll_y * m_to_pixels_coeff);
	map_ptr->m_to_pixels_coeff = m_to_pixels_coeff;

	return(n);
}

/*
 *	Delete all map markings on map object.
 */
void SARMenuMapDeleteAllMarkings(sar_menu_map_struct *map_ptr)
{
	if(map_ptr != NULL)
	{
	    int i;
	    sar_menu_map_marking_struct *marking_ptr;


	    for(i = 0; i < map_ptr->total_markings; i++)
	    {
		marking_ptr = map_ptr->marking[i];
		if(marking_ptr == NULL)
		    continue;

		/* Leave icon image data alone (shared). */

		free(marking_ptr);
	    }

	    free(map_ptr->marking);
	    map_ptr->marking = NULL;

	    map_ptr->total_markings = 0;
	}

	return;
}

/*
 *	Draws the map object.
 */
void SARMenuDrawMap(
        gw_display_struct *display,
        sar_menu_struct *m,
        sar_menu_map_struct *map_ptr,
	int map_num
)
{
	int i, width, height;
	int scroll_xp, scroll_yp;
        int	map_wp, map_hp,
		map_xp_cen, map_yp_cen,
		map_xp_min, map_yp_min,
		map_xp_max, map_yp_max;

	int	marking_xp, marking_yp;
	double mtop_coeff;
	sar_image_struct *image;
	v3d_texture_ref_struct *t;
	sar_menu_map_marking_struct *marking_ptr;


        if((display == NULL) ||
           (m == NULL) ||
	   (map_ptr == NULL)
        )
            return;

        width = display->width;
        height = display->height;

	mtop_coeff = map_ptr->m_to_pixels_coeff;

	scroll_xp = map_ptr->scroll_x;
	scroll_yp = map_ptr->scroll_y;

	/* Bounds of map object on menu in pixels. */
	map_xp_cen = (int)(map_ptr->x * width);
	map_yp_cen = (int)(map_ptr->y * height);

	map_wp = (int)(map_ptr->width * width);
	map_hp = (int)(map_ptr->height * height);
	if((map_wp <= 0) ||
           (map_hp <= 0)
	)
	    return;

        map_xp_min = MAX(map_xp_cen - (map_wp / 2), 0);
        map_yp_min = MAX(map_yp_cen - (map_hp / 2), 0);

        map_xp_max = map_xp_min + map_wp;
        map_yp_max = map_yp_min + map_hp;

	/* Convert all coordinates to lower-left origin for easy
	 * OpenGL calculations.
	 */
	i = map_yp_max;
	map_yp_cen = MAX(height - map_yp_cen, 0);
	map_yp_max = MAX(height - map_yp_min, 0);
	map_yp_min = MAX(height - i, 0);


	t = map_ptr->bg_tex;
	if(t != NULL)
	{
	    int tex_wp = (int)MAX(map_ptr->bg_tex_width * mtop_coeff, 10);
            int tex_hp = (int)MAX(map_ptr->bg_tex_height * mtop_coeff, 10);
	    double t_xc_min, t_yc_min, t_xc_max, t_yc_max;

	    t_xc_min = (double)(-scroll_xp + (tex_wp / 2) - (map_wp / 2))
		/ (double)tex_wp;
            t_xc_max = (double)(-scroll_xp + (tex_wp / 2) + (map_wp / 2))
                / (double)tex_wp;

            t_yc_max = (double)(scroll_yp + (tex_hp / 2) - (map_hp / 2))
                / (double)tex_hp;
            t_yc_min = (double)(scroll_yp + (tex_hp / 2) + (map_hp / 2))
                / (double)tex_hp;

	    StateGLEnable(&display->state_gl, GL_TEXTURE_2D);
	    StateGLTexEnvI(
		&display->state_gl, GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE,
		GL_MODULATE
	    );
	    V3DTextureSelect(t);
	    glColor4f(0.5, 0.5, 0.5, 1.0);
            glBegin(GL_QUADS);
            {
		glTexCoord2f((GLfloat)t_xc_min, (GLfloat)t_yc_min);
                glVertex2f((GLfloat)map_xp_min, (GLfloat)map_yp_min);

		glTexCoord2f((GLfloat)t_xc_max, (GLfloat)t_yc_min);
                glVertex2f((GLfloat)map_xp_max, (GLfloat)map_yp_min);

                glTexCoord2f((GLfloat)t_xc_max, (GLfloat)t_yc_max);
                glVertex2f((GLfloat)map_xp_max, (GLfloat)map_yp_max);

                glTexCoord2f((GLfloat)t_xc_min, (GLfloat)t_yc_max);
                glVertex2f((GLfloat)map_xp_min, (GLfloat)map_yp_max);
            }
            glEnd();

	    StateGLDisable(&display->state_gl, GL_TEXTURE_2D);
	}
	else
	{
	    glColor4f(0.0, 0.0, 0.0, 1.0);
	    glBegin(GL_QUADS);
	    {
		glVertex2f((GLfloat)map_xp_min, (GLfloat)map_yp_min);
                glVertex2f((GLfloat)map_xp_max, (GLfloat)map_yp_min);
                glVertex2f((GLfloat)map_xp_max, (GLfloat)map_yp_max);
                glVertex2f((GLfloat)map_xp_min, (GLfloat)map_yp_max);
	    }
	    glEnd();
	}


	/* Begin drawing markings. */
	for(i = 0; i < map_ptr->total_markings; i++)
	{
	    marking_ptr = map_ptr->marking[i];
	    if(marking_ptr == NULL)
		continue;

	    /* Calculate position. */
	    marking_xp = (int)((marking_ptr->x * mtop_coeff) +
		map_xp_cen + scroll_xp);
            marking_yp = (int)(marking_ptr->y * mtop_coeff +
		map_yp_cen + scroll_yp);

	    /* Handle by type. */
	    switch(marking_ptr->type)
	    {
	      /* ***************************************************** */
	      case SAR_MENU_MAP_MARKING_TYPE_ICON:
		image = marking_ptr->icon;
		if(image != NULL)
		{
		    int	img_hw = image->width / 2,
			img_hh = image->height / 2;

		    /* Out of bounds? */
		    if((marking_xp <= (map_xp_min + img_hw)) ||
		       (marking_yp <= (map_yp_min + img_hh)) ||
		       (marking_xp >= (map_xp_max - img_hw)) ||
		       (marking_yp >= (map_yp_max - img_hh))
		    )
			break;

                    glPixelZoom(1.0, -1.0);
		    StateGLEnable(&display->state_gl, GL_ALPHA_TEST);
                    StateGLAlphaFunc(&display->state_gl, GL_GREATER, 0.5);
		    glRasterPos2i(
		        marking_xp - (image->width / 2),
		        marking_yp + (image->height / 2)
		    );
                    glDrawPixels(
                        image->width, image->height,
                        GL_RGBA, GL_UNSIGNED_BYTE,
                        image->data
                    );
		}
		break;

	      case SAR_MENU_MAP_MARKING_TYPE_INTERCEPT_LINE:
/* Need better bounds checking. */
		if(marking_xp <= map_xp_min)
		    marking_xp = map_xp_min + 1;
                else if(marking_xp >= map_xp_max)
                    marking_xp = map_xp_max - 1;

                if(marking_yp <= map_yp_min)
                    marking_yp = map_yp_min + 1;
                else if(marking_yp >= map_yp_max)
                    marking_yp = map_yp_max - 1;


		glColor4f(
		    (GLfloat)marking_ptr->fg_color.r,
		    (GLfloat)marking_ptr->fg_color.g,
		    (GLfloat)marking_ptr->fg_color.b,
		    (GLfloat)marking_ptr->fg_color.a
		);
		glBegin(GL_LINES);
		{
		    int marking_x2p = (int)((marking_ptr->x_end * mtop_coeff) +
			map_xp_cen + scroll_xp),
			marking_y2p = (int)((marking_ptr->y_end * mtop_coeff) +
			map_yp_cen + scroll_yp);

/* Need better bounds checking. */
                    if(marking_x2p <= map_xp_min)
                        marking_x2p = map_xp_min + 1;
                    else if(marking_x2p >= map_xp_max)
                        marking_x2p = map_xp_max - 1;

                    if(marking_y2p <= map_yp_min)
                        marking_y2p = map_yp_min + 1;
                    else if(marking_y2p >= map_yp_max)
                        marking_y2p = map_yp_max - 1;

		    glVertex2f((GLfloat)marking_xp, (GLfloat)marking_yp);
		    glVertex2f((GLfloat)marking_x2p, (GLfloat)marking_y2p);
		}
		glEnd();
		break;
	    }
	}

	return;
}
