#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <math.h>

#include "../include/os.h"

#ifdef __MSW__
# include <windows.h>
#endif

#include <GL/gl.h>

#include "../include/string.h"

#include "sfm.h"

#include "sarreality.h"
#include "obj.h"
#include "sound.h"
#include "sar.h"


void SARSetDirectionRadians(
	sar_direction_struct *dir,
	double heading, double pitch, double bank
);

int SARObjIsAllocated(
	sar_object_struct **ptr, int total,
        int n
);
int SARGetObjectNumberFromPointer(
	sar_scene_struct *scene,
	sar_object_struct **ptr, int total,
        sar_object_struct *obj_ptr
);
int SARIsTextureAllocated(sar_scene_struct *scene, int n);
sar_object_struct *SARObjMatchPointerByName(
	sar_scene_struct *scene,
	sar_object_struct **ptr, int total,
        const char *name, int *obj_num
);

v3d_texture_ref_struct *SARGetTextureRefByName(
	sar_scene_struct *scene, const char *name
);
int SARGetTextureRefNumberByName(
	sar_scene_struct *scene, const char *name
);

sar_visual_model_struct *SARVisualModelNew(
	sar_scene_struct *scene,
	const char *filename, const char *name
);
void *SARVisualModelNewList(sar_visual_model_struct *vmodel);
int SARVisualModelGetRefCount(sar_visual_model_struct *vmodel);  
void SARVisualModelRef(sar_visual_model_struct *vmodel);
void SARVisualModelUnref(sar_visual_model_struct *vmodel);
void SARVisualModelCallList(sar_visual_model_struct *vmodel);
void SARVisualModelDeleteAll(sar_scene_struct *scene);

sar_cloud_layer_struct *SARCloudLayerCreate(
	sar_scene_struct *scene,
	int tile_width, int tile_height,
        double close_range, double altitude,
        const char *tex_name
);
void SARCloudLayerDestroy(sar_cloud_layer_struct *cloud_layer_ptr);

int SARObjAddToGroundList(
	sar_scene_struct *scene, sar_object_struct *obj_ptr
);
void SARObjRemoveFromGroundList(
	sar_scene_struct *scene, sar_object_struct *obj_ptr
);

int SARObjAddToHumanRescueList(
	sar_scene_struct *scene, sar_object_struct *obj_ptr
);
void SARObjRemoveFromHumanRescueList(
        sar_scene_struct *scene, sar_object_struct *obj_ptr
);

int SARObjAddContactBoundsSpherical(
	sar_object_struct *obj_ptr,
	sar_obj_flags_t crash_flags, int crash_type,
	double contact_radius
);
int SARObjAddContactBoundsCylendrical(
        sar_object_struct *obj_ptr,
        sar_obj_flags_t crash_flags, int crash_type,
        double contact_radius,
	double contact_h_min, double contact_h_max
);
int SARObjAddContactBoundsRectangular(
        sar_object_struct *obj_ptr,
        sar_obj_flags_t crash_flags, int crash_type,
	double contact_x_min, double contact_x_max,
	double contact_y_min, double contact_y_max,
	double contact_z_min, double contact_z_max
);

int SARObjCreateIntercept(
        sar_scene_struct *scene,
        sar_obj_intercept_struct ***ptr, int *total,
        sar_obj_flags_t flags,
        double x, double y, double z,
        double radius,
        double urgency,
        const char *name
);
int SARObjCreateLandingGear(
        sar_scene_struct *scene,
        sar_obj_landing_gear_struct ***ptr, int *total
);
int SARObjCreateRotor(
	sar_scene_struct *scene,
	sar_obj_rotor_struct ***ptr, int *total
);
int SARObjCreate(
	sar_scene_struct *scene,
	sar_object_struct ***ptr, int *total,
        int type
);

void SARObjDeleteIntercepts(
	sar_scene_struct *scene,
	sar_obj_intercept_struct ***ptr, int *total
);
void SARObjDeleteLights(
	sar_scene_struct *scene,
	sar_light_struct ***ptr, int *total
);
void SARObjDeleteLandingGears(
	sar_scene_struct *scene,
	sar_obj_landing_gear_struct ***ptr, int *total
);
void SARObjDeleteRotors(
	sar_scene_struct *scene,
	sar_obj_rotor_struct ***ptr, int *total
);
void SARObjDeleteDoors(
        sar_scene_struct *scene,
        sar_obj_door_struct ***ptr, int *total
);
void SARObjDelete(
	void *core_ptr,
	sar_object_struct ***ptr, int *total, int n
);

void SARObjGenerateTilePlane(
	double min, double max,
	double tile_width, double tile_height
);


#define MAX(a,b)        (((a) > (b)) ? (a) : (b))
#define MIN(a,b)        (((a) < (b)) ? (a) : (b))
#define CLIP(a,l,h)     (MIN(MAX((a),(l)),(h)))


/*
 *	Sets the given attitude values in radians to all members
 *	of the given direction structure.
 */
void SARSetDirectionRadians(
        sar_direction_struct *dir,
        double heading, double pitch, double bank
)
{
	if(dir == NULL)
	    return;


	dir->heading = SFMSanitizeRadians(heading);
	dir->pitch = SFMSanitizeRadians(pitch);
	dir->bank = SFMSanitizeRadians(bank);

	return;
}

/*
 *	Returns true if object n is allocated in the given array.
 */
int SARObjIsAllocated(
	sar_object_struct **ptr, int total,
	int n
)
{
	if((ptr == NULL) ||
           (n < 0) ||
           (n >= total)
	)
	    return(0);
	else if(ptr[n] == NULL)
	    return(0);
	else
	    return(1);
}

/*
 *	Returns the number of the object pointer in the given pointer
 *	array or -1 if no match.
 */
int SARGetObjectNumberFromPointer(
	sar_scene_struct *scene,
	sar_object_struct **ptr, int total,
	sar_object_struct *obj_ptr
)
{
	int i;

	for(i = 0; i < total; i++)
	{
	    if(ptr[i] == obj_ptr)
		return(i);
	}

	return(-1);
}

/*
 *	Checks if the texture is allocated on the scene.
 */
int SARIsTextureAllocated(
	sar_scene_struct *scene,
	int n
)
{
	if(scene == NULL)
	    return(0);
	else if((n < 0) || (n >= scene->total_texture_refs))
	    return(0);
	else if(scene->texture_ref[n] == NULL)
	    return(0);
	else
	    return(1);
}

/*
 *	Returns the pointer to the object matching the given name or
 *	NULL if no match was made.
 */
sar_object_struct *SARObjMatchPointerByName(
	sar_scene_struct *scene,
        sar_object_struct **ptr, int total,
        const char *name, int *obj_num
) 
{
	int i;
	sar_object_struct *obj_ptr;


	/* Reset object number return value. */
	if(obj_num != NULL)
	    (*obj_num) = -1;

	/* Go through each object. */
	for(i = 0; i < total; i++)
	{
	    obj_ptr = ptr[i];
	    if(obj_ptr == NULL)
		continue;

	    if(obj_ptr->name == NULL)
		continue;

	    /* Names match? */
	    if(!strcasecmp(obj_ptr->name, name))
	    {
		if(obj_num != NULL)
		    (*obj_num) = i;

		return(obj_ptr);
	    }
	}

	return(NULL);
}

/*
 *	Returns a pointer to the texture referance structure matching
 *	the given name. Can return NULL for failed match.
 */
v3d_texture_ref_struct *SARGetTextureRefByName(
        sar_scene_struct *scene, const char *name
)
{
	int i, total;
	v3d_texture_ref_struct *t, **ptr;


	if((scene == NULL) ||
           (name == NULL)
	)
	    return(NULL);


	ptr = scene->texture_ref;
	total = scene->total_texture_refs;

	for(i = 0; i < total; i++)
	{
	    t = ptr[i];
	    if(t == NULL)
		continue;

	    if(t->name == NULL)
		continue;

	    if(!strcasecmp(t->name, name))
		return(t);
	}

	return(NULL);
}

/*
 *	Same as SARGetTextureRefByName() except that it returns
 *	the index number of the texture or -1 on no match.
 */
int SARGetTextureRefNumberByName(
	sar_scene_struct *scene, const char *name
)
{
        int i, total;
        v3d_texture_ref_struct *t, **ptr;


        if((scene == NULL) ||
           (name == NULL) 
        )
            return(-1);

        ptr = scene->texture_ref;
        total = scene->total_texture_refs;

        for(i = 0; i < total; i++)
        {
            t = ptr[i];
            if(t == NULL)
                continue;

            if(t->name == NULL)
                continue;

            if(!strcasecmp(t->name, name))
                return(i);
        }

        return(-1);
}


/*
 *	Creates a new visual model structure on the given scene structure
 *	or returns an existing one if filename and name both match the
 *	name of an existing visual model on the scene structure.
 *
 *	If filename or name is NULL then a new visual model will be
 *	created and will be treated as `unique' meaning that it will not
 *	be returned for match with any other call to SARVisualModelNew()
 *	(since the filename and name are unmatchable).
 */
sar_visual_model_struct *SARVisualModelNew(
        sar_scene_struct *scene,
        const char *filename, const char *name
)
{
	int i;
	sar_visual_model_struct *vmodel;

	if(scene == NULL)
	    return(NULL);

	/* Both filename and name given? */
	if((filename != NULL) && (name != NULL))
	{
	    /* Check for existing visual model structure on scene
	     * structure.
	     */
	    for(i = 0; i < scene->total_visual_models; i++)
	    {
		vmodel = scene->visual_model[i];
		if(vmodel == NULL)
		    continue;

		/* Visual model on scene structure has both filename
		 * and name set?
		 */
		if((vmodel->filename != NULL) && (vmodel->name != NULL))
		{
		    /* Given filename and name matches? */
		    if(!strcmp(vmodel->filename, filename) &&
                       !strcmp(vmodel->name, name)
		    )
		    {
/*
fprintf(stderr,
"SARVisualModelNew(): Returning existing model for `%s'\n",
 name
);
 */
			return(vmodel);
		    }
		}
	    }
	}

	/* Need to create new visual model on scene if this is reached. */

	if(scene->total_visual_models < 0)
	    scene->total_visual_models = 0;

	/* Allocate more pointers. */
	i = scene->total_visual_models;
	scene->total_visual_models = i + 1;
	scene->visual_model = (sar_visual_model_struct **)realloc(
	    scene->visual_model,
	    scene->total_visual_models * sizeof(sar_visual_model_struct *)
	);
	if(scene->visual_model == NULL)
	{
	    scene->total_visual_models = 0;
	    return(NULL);
	}

	/* Allocate new structure. */
	vmodel = (sar_visual_model_struct *)calloc(
	    1, sizeof(sar_visual_model_struct)
	);
	scene->visual_model[i] = vmodel;
	if(vmodel != NULL)
	{
	    vmodel->load_state = SAR_VISUAL_MODEL_NOT_LOADED;

	    vmodel->ref_count = 0;

	    if(filename != NULL)
		vmodel->filename = strdup(filename);
	    if(name != NULL)
		vmodel->name = strdup(name);

	    vmodel->data = NULL;
	}

	return(vmodel);
}

/*
 *	(Re)generates a list on the given visual model, can return NULL
 *	on failure.
 *
 *	The return value is really a GLint (not a void *).
 */
void *SARVisualModelNewList(sar_visual_model_struct *vmodel)
{
	GLint list;

	if(vmodel == NULL)
	    return(NULL);

	/* Get existing GL list if any. */
	list = (GLint)vmodel->data;
	if(list > 0)
	{
	    /* Already has an allocated GL list for some odd reason,
	     * oh well, delete it and then create a new GL list.
	     */
	    glDeleteLists(list, 1);
	}

	/* Create new GL list. */
	list = glGenLists(1);

	/* Set new GL list as the data pointer on the visual model
	 * structure.
	 */
	vmodel->data = (void *)list;

	return((void *)list);
}

/*
 *	Returns number of ref counts on visual model.
 *
 *	Can return -1 if visual model is NULL.
 */
int SARVisualModelGetRefCount(sar_visual_model_struct *vmodel)
{
        if(vmodel != NULL)
            return(vmodel->ref_count);
	else
	    return(-1);
} 

/*
 *	Increase ref count on visual model by one.
 */
void SARVisualModelRef(sar_visual_model_struct *vmodel)
{
	if(vmodel != NULL)
	    vmodel->ref_count++;
	return;
}

/*
 *	Decrease ref count on visual model by one.
 */
void SARVisualModelUnref(sar_visual_model_struct *vmodel)
{
	if(vmodel != NULL)
	{
	    /* No existing referance counts to unref? */
	    if(vmodel->ref_count < 1)
	    {
		/* Print warning. */
		fprintf(
		    stderr,
"SARVisualModelUnref(): Visual model `%s' has no ref counts to decrease.\n",
vmodel->name
		);
	    }
	    else
	    {
		/* Reduce referance count by 1. */
		vmodel->ref_count--;
	    }
	    /* Sanitize referance count. */
	    if(vmodel->ref_count < 0)
	        vmodel->ref_count = 0;

	    /* Has referance count reached 0? */
	    if(vmodel->ref_count == 0)
	    {
/* Should we delete the visual model if the ref count drops back to 0? */


	    }
	}

	return;
}

/*
 *	Calls the GL list on the visual model if it is not 0.
 */
void SARVisualModelCallList(sar_visual_model_struct *vmodel)
{
	GLint list;

	if(vmodel == NULL)
	    return;

	list = (GLint)vmodel->data;
	if(list > 0)
	    glCallList(list);

	return;
}

/*
 *	Dealloates all the visual models on the given scene structure
 *	and deletes the GL list on each visual model structure.
 */
void SARVisualModelDeleteAll(sar_scene_struct *scene)
{
	int i;
	GLint list;
	sar_visual_model_struct *vmodel;


	if(scene == NULL)
	    return;

	for(i = 0; i < scene->total_visual_models; i++)
	{
	    vmodel = scene->visual_model[i];
	    if(vmodel == NULL)
		continue;

if(vmodel->ref_count > 0)
 fprintf(stderr,
  "SARVisualModelDeleteAll(): Visual model `%s' still has %i ref_counts.\n",
  vmodel->name, vmodel->ref_count
 );

	    /* Free filename. */
            free(vmodel->filename);
            vmodel->filename = NULL;

	    /* Free name. */
	    free(vmodel->name);
	    vmodel->name = NULL;

	    /* Delete the GL list. */
	    list = (GLint)vmodel->data;
	    if(list > 0)
		glDeleteLists(list, 1);

	    /* Deallocate visual model structure itself. */
	    free(vmodel);
	    scene->visual_model[i] = NULL;
	}

	free(scene->visual_model);
	scene->visual_model = NULL;
	scene->total_visual_models = 0;

	return;
}


/*
 *	Allocates a new cloud layer structure.
 */
sar_cloud_layer_struct *SARCloudLayerCreate(
	sar_scene_struct *scene,
	int tile_width, int tile_height,
	double close_range,
	double altitude,
	const char *tex_name
)
{
	GLint list;
	double min, max;
	v3d_texture_ref_struct *t;
	sar_visual_model_struct **vmodel;
	sar_cloud_layer_struct *cloud_layer_ptr = (sar_cloud_layer_struct *)
	    calloc(1, sizeof(sar_cloud_layer_struct));


	if(cloud_layer_ptr == NULL)
	    return(NULL);

	cloud_layer_ptr->tile_width = tile_width;
	cloud_layer_ptr->tile_height = tile_height;
	cloud_layer_ptr->close_range = close_range;
	cloud_layer_ptr->z = altitude;

	/* Match texture by name. */
	t = SARGetTextureRefByName(scene, tex_name);

	/* Calculate entire size limits. */
	min = -cloud_layer_ptr->close_range;
	max = cloud_layer_ptr->close_range;

	/* Get pointer to visual model. */
        vmodel = &cloud_layer_ptr->visual_model;

	/* Unref visual model if its pointer to referance is not NULL. */
	if((*vmodel) != NULL)
	{
	    SARVisualModelUnref(*vmodel);
	    (*vmodel) = NULL;
	}

	/* Create new visual model. */
	(*vmodel) = SARVisualModelNew(
	    scene, NULL, NULL
	);
	/* Increment referance count so it should now be 1. */
	SARVisualModelRef(*vmodel);

	/* (Re)generate GL list on visual model structure. */
	list = (GLint)SARVisualModelNewList(*vmodel);
	if(list != 0)
	{
	    /* Mark as loading. */
	    (*vmodel)->load_state = SAR_VISUAL_MODEL_LOADING;

	    /* Begin recording GL list. */
	    glNewList(list, GL_COMPILE);
	    {
		/* Set shading type. */
		glShadeModel(GL_FLAT);

		/* Select texture if defined. */
		V3DTextureSelect(t);

		/* Do not set color. */

		/* Close range textured tiles. */
		SARObjGenerateTilePlane(
		    min, max,			/* Min and max. */
		    tile_width, tile_height	/* Tiling size. */
		);
	    }
	    glEndList();

            /* Mark as done loading. */
            (*vmodel)->load_state = SAR_VISUAL_MODEL_LOADED;
	}

	return(cloud_layer_ptr);
}

/*
 *	Deallocates all substructures of the cloud layer and
 *	the cloud layer structure itself.
 */
void SARCloudLayerDestroy(
	sar_cloud_layer_struct *cloud_layer_ptr
)
{
	if(cloud_layer_ptr != NULL)
	{
	    SARVisualModelUnref(cloud_layer_ptr->visual_model);
	    cloud_layer_ptr->visual_model = NULL;

	    free(cloud_layer_ptr->tex_name);
	    free(cloud_layer_ptr);
	}

	return;
}

/*
 *	Checks if the object is in the scene's ground list.
 *	If not, then it will add it to the list. Returns 0 on success
 *	or negative on error/already in list.
 */
int SARObjAddToGroundList(
	sar_scene_struct *scene,  
        sar_object_struct *obj_ptr
)
{
	int i;


	if((scene == NULL) ||
           (obj_ptr == NULL)
	)
	    return(-1);

	if(scene->total_ground_objects < 0)
	    scene->total_ground_objects = 0;

	/* Check if already in list. */
	for(i = 0; i < scene->total_ground_objects; i++)
	{
	    if(obj_ptr == scene->ground_object[i])
		return(-2);
	}

	/* Not in list, so add to list. */
        for(i = 0; i < scene->total_ground_objects; i++)
        {
            if(scene->ground_object[i] == NULL)
            {
		/* Found an empty entry. */
		scene->ground_object[i] = obj_ptr;
		return(0);
	    }
        }

	/* Need to allocate more entries. */
	i = scene->total_ground_objects;
	scene->total_ground_objects++;

	scene->ground_object = (sar_object_struct **)realloc(
	    scene->ground_object,
	    scene->total_ground_objects * sizeof(sar_object_struct *)
	);
	if(scene->ground_object == NULL)
	{
	    scene->total_ground_objects = 0;
	    return(-1);
	}

	scene->ground_object[i] = obj_ptr;

	return(0);
}

/*
 *	Removes the object from the ground list if it is in there.
 */
void SARObjRemoveFromGroundList(
	sar_scene_struct *scene,
	sar_object_struct *obj_ptr
)
{
	int i;


        if((scene == NULL) ||
           (obj_ptr == NULL)
        )
            return;

        for(i = 0; i < scene->total_ground_objects; i++)
        {
            if(obj_ptr == scene->ground_object[i])
                scene->ground_object[i] = NULL;
        }

	return;
}

/*
 *	Adds object pointer to the humans need rescue list on the scene:
 */
int SARObjAddToHumanRescueList(
	sar_scene_struct *scene,
        sar_object_struct *obj_ptr
)
{
        int i;
            
            
        if((scene == NULL) ||
           (obj_ptr == NULL)
        )
            return(-1);
            
        if(scene->total_human_need_rescue_objects < 0)
            scene->total_human_need_rescue_objects = 0;

        /* Check if already in list. */
        for(i = 0; i < scene->total_human_need_rescue_objects; i++)
        {
            if(obj_ptr == scene->human_need_rescue_object[i])
                return(-2);
        }

        /* Not in list, so add to list. */
        for(i = 0; i < scene->total_human_need_rescue_objects; i++)
        {
            if(scene->human_need_rescue_object[i] == NULL)
            {
                /* Found an empty entry. */
                scene->human_need_rescue_object[i] = obj_ptr;
                return(0);
            }
        }

        /* Need to allocate more entries. */
        i = scene->total_human_need_rescue_objects;
        scene->total_human_need_rescue_objects++;

        scene->human_need_rescue_object = (sar_object_struct **)realloc(
            scene->human_need_rescue_object,
            scene->total_human_need_rescue_objects * sizeof(sar_object_struct *)
        );
        if(scene->human_need_rescue_object == NULL)
        {
            scene->total_human_need_rescue_objects = 0;
            return(-1);
        }

        scene->human_need_rescue_object[i] = obj_ptr;

        return(0);
}

/*
 *	Removes object from humans that need rescue list if it is in
 *	there.
 */
void SARObjRemoveFromHumanRescueList(
	sar_scene_struct *scene,
	sar_object_struct *obj_ptr
)
{
        int i;

        if((scene == NULL) ||   
           (obj_ptr == NULL)
        )
            return;

        for(i = 0; i < scene->total_human_need_rescue_objects; i++)
        {
            if(obj_ptr == scene->human_need_rescue_object[i])
                scene->human_need_rescue_object[i] = NULL;
        }

        return;
}


/*
 *	Allocate (as needed) contact bounds structure to object
 *	for shape SAR_CONTACT_SHAPE_SPHERICAL.
 *
 *      Returns non-zero on error.
 */
int SARObjAddContactBoundsSpherical(
        sar_object_struct *obj_ptr,
	sar_obj_flags_t crash_flags, int crash_type,
        double contact_radius
)
{
	sar_contact_bounds_struct *cb;


	if(obj_ptr == NULL)
	    return(-1);

	/* Allocate as needed. */
	if(obj_ptr->contact_bounds == NULL)
	    obj_ptr->contact_bounds = (sar_contact_bounds_struct *)
		calloc(1, sizeof(sar_contact_bounds_struct));

	cb = obj_ptr->contact_bounds;
	if(cb == NULL)
	    return(-1);

	cb->crash_flags = crash_flags;
	cb->crash_type = crash_type;
	cb->contact_shape = SAR_CONTACT_SHAPE_SPHERICAL;

	cb->contact_radius = MAX(contact_radius, 0.0);

	return(0);
}

/*
 *      Allocate (as needed) contact bounds structure to object
 *      for shape SAR_CONTACT_SHAPE_CYLENDRICAL.
 *
 *      Returns non-zero on error.
 */
int SARObjAddContactBoundsCylendrical(
        sar_object_struct *obj_ptr,
	sar_obj_flags_t crash_flags, int crash_type,
        double contact_radius,
        double contact_h_min, double contact_h_max 
)
{
        sar_contact_bounds_struct *cb;


        if(obj_ptr == NULL)
            return(-1);

        /* Allocate as needed. */
        if(obj_ptr->contact_bounds == NULL)
            obj_ptr->contact_bounds = (sar_contact_bounds_struct *)
                calloc(1, sizeof(sar_contact_bounds_struct));

        cb = obj_ptr->contact_bounds;
        if(cb == NULL)
            return(-1);

        cb->crash_flags = crash_flags;
        cb->crash_type = crash_type;
        cb->contact_shape = SAR_CONTACT_SHAPE_CYLENDRICAL;

        cb->contact_radius = MAX(contact_radius, 0.0);
	cb->contact_h_min = contact_h_min;
        cb->contact_h_max = contact_h_max;

	/* Flip height as needed. */
	if(cb->contact_h_min > cb->contact_h_max)
	{
	    double h = cb->contact_h_min;

	    cb->contact_h_min = cb->contact_h_max;
	    cb->contact_h_max = h;
	}

	return(0);
}

/*
 *      Allocate (as needed) contact bounds structure to object
 *      for shape SAR_CONTACT_SHAPE_RECTANGULAR.
 *
 *	Returns non-zero on error.
 */
int SARObjAddContactBoundsRectangular(
        sar_object_struct *obj_ptr,
	sar_obj_flags_t crash_flags, int crash_type,
        double contact_x_min, double contact_x_max,
        double contact_y_min, double contact_y_max,
        double contact_z_min, double contact_z_max 
)
{
        sar_contact_bounds_struct *cb;


        if(obj_ptr == NULL)
            return(-1);

        /* Allocate as needed. */
        if(obj_ptr->contact_bounds == NULL)
            obj_ptr->contact_bounds = (sar_contact_bounds_struct *)
                calloc(1, sizeof(sar_contact_bounds_struct));

        cb = obj_ptr->contact_bounds;
        if(cb == NULL)
            return(-1);

        cb->crash_flags = crash_flags;
        cb->crash_type = crash_type;
        cb->contact_shape = SAR_CONTACT_SHAPE_RECTANGULAR;

        cb->contact_x_min = contact_x_min;
	cb->contact_x_max = contact_x_max;
        if(cb->contact_x_min > cb->contact_x_max)
        {
            double d = cb->contact_x_min;

            cb->contact_x_min = cb->contact_x_max;
            cb->contact_x_max = d;
        }

	cb->contact_y_min = contact_y_min;
	cb->contact_y_max = contact_y_max;
        if(cb->contact_y_min > cb->contact_y_max)
        {
            double d = cb->contact_y_min;

            cb->contact_y_min = cb->contact_y_max;
            cb->contact_y_max = d;
        }

	cb->contact_z_min = contact_z_min;
	cb->contact_z_max = contact_z_max;
        if(cb->contact_z_min > cb->contact_z_max)
        {
            double d = cb->contact_z_min;

            cb->contact_z_min = cb->contact_z_max;
            cb->contact_z_max = d;
        }

	/* Calculate trig values based on object's current heading. */
	cb->cos_heading = cos(-obj_ptr->dir.heading);
        cb->sin_heading = sin(-obj_ptr->dir.heading);

        return(0);
}


/*
 *	Adds an intercept to the object. Returns the allocated intercept
 *	index number or -1 on error.
 */
int SARObjCreateIntercept(
	sar_scene_struct *scene,
	sar_obj_intercept_struct ***ptr, int *total,
	sar_obj_flags_t flags,
	double x, double y, double z,
	double radius,
	double urgency,
	const char *name
)
{
	int n;
	sar_obj_intercept_struct *intercept_ptr;


	if((ptr == NULL) || (total == NULL))
	    return(-1);

	/* Increase total. */
	if((*total) < 0)
	    (*total) = 0;
	n = (*total);
	(*total) = n + 1;

	/* Allocate more pointers. */
	(*ptr) = (sar_obj_intercept_struct **)realloc(
	    *ptr,
	    (*total) * sizeof(sar_obj_intercept_struct *)
	);
	if((*ptr) == NULL)
	{
	    (*total) = 0;
	    return(-1);
	}

	/* Allocate structure. */
	intercept_ptr = (sar_obj_intercept_struct *)calloc(
	    1,
	    sizeof(sar_obj_intercept_struct)
	);
	(*ptr)[n] = intercept_ptr;
	if(intercept_ptr == NULL)
	    return(-1);

	/* Set values. */
	intercept_ptr->flags = flags;
	intercept_ptr->x = x;
	intercept_ptr->y = y;
	intercept_ptr->z = z;
	intercept_ptr->radius = MAX(radius, 0.0);
	intercept_ptr->urgency = CLIP(urgency, 0.0, 1.0);
	intercept_ptr->name = ((name == NULL) ? NULL : strdup(name));

	return(n);
}

/*
 *	Allocates a new landing gear structure, returning
 *	the new landing gear index number or -1 on error.
 */
int SARObjCreateLandingGear(
	sar_scene_struct *scene,
	sar_obj_landing_gear_struct ***ptr, int *total
)
{
	int i, n;
	sar_obj_landing_gear_struct *lgear_ptr;


	if((ptr == NULL) ||
           (total == NULL)
	)
	    return(-1);

	if(*total < 0)
	    *total = 0;

	for(i = 0; i < *total; i++)
	{
	    if((*ptr)[i] == NULL)
		break;
	}
	if(i < *total)
	{
	    n = i;
	}
	else
	{
	    n = *total;
	    *total = *total + 1;

	    *ptr = (sar_obj_landing_gear_struct **)realloc(
		*ptr,
		*total * sizeof(sar_obj_landing_gear_struct *)
	    );
	    if(*ptr == NULL)
	    {
		*total = 0;
		return(-1);
	    }
	}

	(*ptr)[n] = (sar_obj_landing_gear_struct *)calloc(
	    1,
	    sizeof(sar_obj_landing_gear_struct)
	);
	lgear_ptr = (*ptr)[n];
	if(lgear_ptr == NULL)
	{
	    return(-1);
	}

	/* Reset values. */
	lgear_ptr->flags = 0;

	lgear_ptr->extended_dir.heading = 0;
	lgear_ptr->extended_dir.pitch = (0.5 * PI);
	lgear_ptr->extended_dir.bank = 0;

	lgear_ptr->anim_pos = (sar_grad_anim_t)-1;
	lgear_ptr->anim_rate = (sar_grad_anim_t)(3 * SAR_SEC_TO_CYCLE_COEFF *
	    (sar_grad_anim_t)-1
	);

	lgear_ptr->visual_model = NULL;

	return(n);
}

/*
 *      Allocates a new rotor structure, returning
 *      the new rotor index number or -1 on error.
 */
int SARObjCreateRotor(
        sar_scene_struct *scene,
        sar_obj_rotor_struct ***ptr, int *total
) 
{
        int i, n;
        sar_obj_rotor_struct *rotor_ptr;


        if((ptr == NULL) ||
           (total == NULL)
        )
            return(-1);

        if(*total < 0)
            *total = 0;

        for(i = 0; i < *total; i++)
        {
            if((*ptr)[i] == NULL)
                break;
        }
        if(i < *total)
        {
            n = i;
        }
        else
        {
            n = *total;
            *total = *total + 1;

            *ptr = (sar_obj_rotor_struct **)realloc(
                *ptr,
                *total * sizeof(sar_obj_rotor_struct *)
           );
            if(*ptr == NULL) 
            {
                *total = 0;
                return(-1);
            }
        }

        (*ptr)[n] = (sar_obj_rotor_struct *)calloc( 
            1,
            sizeof(sar_obj_rotor_struct)
        );
        rotor_ptr = (*ptr)[n];
        if(rotor_ptr == NULL)
        {
            return(-1);
        }

        /* Reset values. */
        rotor_ptr->flags = 0;
 
        rotor_ptr->dir.heading = 0;
	rotor_ptr->dir.pitch = (0.5 * PI);
	rotor_ptr->dir.bank = 0;

        rotor_ptr->anim_pos = (sar_grad_anim_t)-1;

        rotor_ptr->visual_model = NULL;

        return(n);
}



/*
 *	Creates a new object of the specified type, returning its
 *	index number in the given array or -1 on error.
 *
 *	The object structure's member `data' will be allocated as
 *	well determined by its type.
 */
int SARObjCreate(
	sar_scene_struct *scene,
	sar_object_struct ***ptr, int *total,
	int type
)
{
	int i, n, len;
	sar_object_struct *obj_ptr;


	if((scene == NULL) ||
           (ptr == NULL) ||
           (total == NULL) ||
	   (type <= SAR_OBJ_TYPE_GARBAGE)
	)
	    return(-1);

	if(*total < 0)
	    *total = 0;

	for(i = 0; i < *total; i++)
	{
	    if((*ptr)[i] == NULL)
		break;

	    if((*ptr)[i]->type <= SAR_OBJ_TYPE_GARBAGE)
		break;
	}
	if(i < *total)
	{
	    /* Got already allocated structure. */
	    n = i;
	}
	else
	{
	    /* Need to allocate a new object structure. */
	    n = *total;
	    *total = *total + 1;

	    *ptr = (sar_object_struct **)realloc(
		*ptr,
		*total * sizeof(sar_object_struct *)
	    );
	    if(*ptr == NULL)
	    {
		*total = 0;
		return(-1);
	    }

	    (*ptr)[n] = NULL;
	}

	/* Allocate object structure as needed. */
	if((*ptr)[n] == NULL)
	{
	    (*ptr)[n] = (sar_object_struct *)calloc(
		1,
		sizeof(sar_object_struct)
	    );
	    if((*ptr)[n] == NULL)
	    {
		return(-1);
	    }
	}

	obj_ptr = (*ptr)[n];


	/* Set object type. */
	obj_ptr->type = type;

	/* Mark birth time. */
	obj_ptr->birth_time_ms = cur_millitime;
	obj_ptr->birth_time_sec = cur_systime;

	/* Allocate substructure (if len is 0 then implies has no
	 * substructure).
	 */
	switch(type)
	{
	  case SAR_OBJ_TYPE_GARBAGE:
            len = 0;
	    break;

	  case SAR_OBJ_TYPE_STATIC:
            len = 0;
	    break;

	  case SAR_OBJ_TYPE_AUTOMOBILE:
            len = 0;
	    break;

	  case SAR_OBJ_TYPE_WATERCRAFT:
	    len = 0;
	    break;

	  case SAR_OBJ_TYPE_AIRCRAFT:
	    len = sizeof(sar_object_aircraft_struct);
	    break;

	  case SAR_OBJ_TYPE_GROUND:
	    len = sizeof(sar_object_ground_struct);
	    break;

	  case SAR_OBJ_TYPE_RUNWAY:
	    len = sizeof(sar_object_runway_struct);
	    break;

	  case SAR_OBJ_TYPE_HELIPAD:
	    len = sizeof(sar_object_helipad_struct);
            break;

	  case SAR_OBJ_TYPE_HUMAN:
	    len = sizeof(sar_object_human_struct);
	    break;


	  case SAR_OBJ_TYPE_SMOKE:
            len = sizeof(sar_object_smoke_struct);
            break;

          case SAR_OBJ_TYPE_FIRE:
            len = 0;
            break;

          case SAR_OBJ_TYPE_EXPLOSION:
            len = sizeof(sar_object_explosion_struct);
            break;

	  case SAR_OBJ_TYPE_PREMODELED:
	    len = sizeof(sar_object_premodeled_struct);
	    break;

	  default:
	    len = 0;
	    break;
	}

	if(len > 0)
	    obj_ptr->data = calloc(1, len);


	/* If object type is SAR_OBJ_TYPE_GROUND, then add to ground
	 * list on scene structure.
	 */
	if((scene != NULL) &&
           (obj_ptr->type == SAR_OBJ_TYPE_GROUND)
	)
	    SARObjAddToGroundList(scene, obj_ptr);


	return(n);
}


/*
 *	Deletes all intercept structures in the given array.
 */
void SARObjDeleteIntercepts(
	sar_scene_struct *scene,
	sar_obj_intercept_struct ***ptr, int *total
)
{
	int i;
	sar_obj_intercept_struct *intercept_ptr;


        if((ptr == NULL) || 
           (total == NULL)
        )
            return;

	for(i = 0; i < (*total); i++)
	{
	    intercept_ptr = (*ptr)[i];
	    if(intercept_ptr == NULL)
		continue;

	    free(intercept_ptr->name);
	    free(intercept_ptr);
	}

	free(*ptr);
	(*ptr) = NULL;
	(*total) = 0;

	return;
}

/*
 *	Deletes all structures and resources in the array of light
 *	structures.
 */
void SARObjDeleteLights(
	sar_scene_struct *scene,
	sar_light_struct ***ptr,
	int *total
)
{
        int i;
        sar_light_struct *light_ptr;


        if((ptr == NULL) ||
           (total == NULL)
        )
            return;

        for(i = 0; i < (*total); i++)
        {
            light_ptr = (*ptr)[i];
            if(light_ptr == NULL)
                continue;

            free(light_ptr);
        }

        free(*ptr);
        (*ptr) = NULL;

        (*total) = 0;

        return;
}

/*
 *	Deletes all structures and resources in the array of landing
 *	gear structures.
 */
void SARObjDeleteLandingGears(
        sar_scene_struct *scene,
	sar_obj_landing_gear_struct ***ptr,
	int *total
)
{
	int i;
	sar_obj_landing_gear_struct *lgear_ptr;


	if((ptr == NULL) ||
           (total == NULL)
	)
	    return;

        for(i = 0; i < (*total); i++)
	{
            lgear_ptr = (*ptr)[i];
	    if(lgear_ptr == NULL)
	        continue;

	    SARVisualModelUnref(lgear_ptr->visual_model);
	    lgear_ptr->visual_model = NULL;

            free(lgear_ptr);
        }

        free(*ptr);
	(*ptr) = NULL;

	(*total) = 0;

	return;
}

/*                  
 *      Deletes all structures and resources in the array of landing
 *      gear structures.
 */                     
void SARObjDeleteRotors(
	sar_scene_struct *scene,
        sar_obj_rotor_struct ***ptr,
        int *total
)                   
{ 
        int i;
        sar_obj_rotor_struct *rotor_ptr;


        if((ptr == NULL) ||
           (total == NULL)
        )
            return;

        for(i = 0; i < *total; i++)
        {
            rotor_ptr = (*ptr)[i];
            if(rotor_ptr == NULL)
                continue;

	    SARVisualModelUnref(rotor_ptr->visual_model);
	    rotor_ptr->visual_model = NULL;

            free(rotor_ptr);
        }

        free(*ptr);
        *ptr = NULL;

        *total = 0;     

        return;
}                     

/*
 *      Deletes all structures and resources in the array of doors.
 */
void SARObjDeleteDoors(
	sar_scene_struct *scene,
	sar_obj_door_struct ***ptr,
	int *total
)
{
        int i;
        sar_obj_door_struct *door_ptr;


        if((ptr == NULL) ||
           (total == NULL)
        )
            return;

        for(i = 0; i < *total; i++)
        {
            door_ptr = (*ptr)[i];
            if(door_ptr == NULL)
                continue;

	    SARVisualModelUnref(door_ptr->visual_model);
	    door_ptr->visual_model = NULL;

            free(door_ptr);
        }

        free(*ptr);
        (*ptr) = NULL;

        (*total) = 0;

	return;
}

/*
 *	Deletes object n from the array, setting it to NULL.
 */
void SARObjDelete(
	void *core_ptr,
        sar_object_struct ***ptr, int *total, int n
)
{
	sar_core_struct *cp = core_ptr;
	sar_scene_struct *scene;
	snd_recorder_struct *recorder;
	char **storage;

	if(cp == NULL)
	    return;

	scene = cp->scene;
	if(scene == NULL)
	{
	    fprintf(stderr,
		"SARObjDelete(): Internal error, scene structure is NULL.\n"
	    );
	    return;
	}

	recorder = cp->recorder;

	if(SARObjIsAllocated(*ptr, *total, n))
	{
	    int i;
	    sar_visual_model_struct **vmodel = NULL;
	    void **sound_play = NULL;
	    sar_object_struct *obj_ptr, *obj_ptr2;
	    sar_object_aircraft_struct *obj_aircraft_ptr;
	    sar_object_runway_struct *obj_runway_ptr;
	    sar_object_helipad_struct *obj_helipad_ptr;
	    sar_object_ground_struct *obj_ground_ptr;
	    sar_object_smoke_struct *obj_smoke_ptr;
	    sar_object_explosion_struct *obj_explosion_ptr;
	    sar_object_human_struct *obj_human_ptr;
	    sar_object_premodeled_struct *obj_premodeled_ptr;
	    sar_obj_door_struct *door_ptr;
	    sar_obj_hoist_struct *hoist_ptr;


	    obj_ptr = (*ptr)[n];

#define DO_UNREF_VISUAL_MODEL	\
if(vmodel != NULL) \
{ \
 if((*vmodel) != NULL) \
 { \
  SARVisualModelUnref(*vmodel); \
  (*vmodel) = NULL; \
 } \
}

#define DO_STOP_SOUND	\
if(sound_play != NULL) \
{ \
 if((*sound_play) != NULL) \
 { \
  SoundStopPlay(recorder, *sound_play); \
  (*sound_play) = NULL; \
 } \
}

#define DO_FREE_STORAGE	\
{ \
 if(storage != NULL) \
 { \
  free(*storage); \
  (*storage) = NULL; \
 } \
}

	    /* Deallocate substructures. */
	    if(obj_ptr->data != NULL)
	    {
		switch(obj_ptr->type)
		{
		  case SAR_OBJ_TYPE_AIRCRAFT:
		    obj_aircraft_ptr = obj_ptr->data;
		    if(obj_aircraft_ptr == NULL)
			break;

		    /* Engine sound play and paths. */
		    sound_play = &obj_aircraft_ptr->engine_inside_sndplay;
		    DO_STOP_SOUND
		    storage = &obj_aircraft_ptr->engine_inside_sndpath;
		    DO_FREE_STORAGE

                    sound_play = &obj_aircraft_ptr->engine_outside_sndplay;
                    DO_STOP_SOUND
		    storage = &obj_aircraft_ptr->engine_outside_sndpath;
		    DO_FREE_STORAGE

                    /* Shadow visual model. */
		    vmodel = &obj_aircraft_ptr->visual_model_shadow;
		    DO_UNREF_VISUAL_MODEL

		    /* Cockpit visual model. */
		    vmodel = &obj_aircraft_ptr->visual_model_cockpit;
		    DO_UNREF_VISUAL_MODEL

		    /* Rotors. */
		    SARObjDeleteRotors(
                        scene,
                        &obj_aircraft_ptr->rotor,
                        &obj_aircraft_ptr->total_rotors
                    );

		    /* Landing gears. */
		    SARObjDeleteLandingGears(
			scene,
			&obj_aircraft_ptr->landing_gear,
			&obj_aircraft_ptr->total_landing_gears
		    );

		    /* Rescue door. */
		    door_ptr = &obj_aircraft_ptr->rescue_door;
                    vmodel = &door_ptr->visual_model;
		    DO_UNREF_VISUAL_MODEL

		    /* Hoist. */
		    hoist_ptr = &obj_aircraft_ptr->hoist;
		    /* Free list of occupants on hoist struct. */
		    free(hoist_ptr->occupant);
		    hoist_ptr->occupant = NULL;
		    hoist_ptr->total_occupants = 0;

		    /* Intercepts. */
		    SARObjDeleteIntercepts(
			scene,
			&obj_aircraft_ptr->intercept,
			&obj_aircraft_ptr->total_intercepts
		    );

		    /* Flight dynamics model. */
		    if((scene->realm != NULL) &&
                       (obj_aircraft_ptr->fdm != NULL)
		    )
		    {
			SFMModelDelete(scene->realm, obj_aircraft_ptr->fdm);
			/* SFM fdm destroy callback will call our callback
			 * function to set this pointer to NULL.
			 */
/*
			obj_aircraft_ptr->fdm = NULL;
 */
		    }
		    break;

		  case SAR_OBJ_TYPE_GROUND:
		    obj_ground_ptr = obj_ptr->data;
                    if(obj_ground_ptr == NULL)
                        break;

		    /* Heightfield z points map. */
		    free(obj_ground_ptr->z_point_value);
		    obj_ground_ptr->z_point_value = NULL;

		    obj_ground_ptr->grid_points_x = 0;
		    obj_ground_ptr->grid_points_y = 0;
		    obj_ground_ptr->grid_points_total = 0;

		    break;

	          case SAR_OBJ_TYPE_RUNWAY:
		    obj_runway_ptr = obj_ptr->data;
		    if(obj_runway_ptr == NULL)
			break;

		    free(obj_runway_ptr->north_label);
		    free(obj_runway_ptr->south_label);

		    break;

                  case SAR_OBJ_TYPE_HELIPAD:
                    obj_helipad_ptr = obj_ptr->data;
                    if(obj_helipad_ptr == NULL)
                        break;

                    free(obj_helipad_ptr->label);

                    break;

                  case SAR_OBJ_TYPE_SMOKE:
                    obj_smoke_ptr = obj_ptr->data;
                    if(obj_smoke_ptr == NULL)
                        break;

                    break;

                  case SAR_OBJ_TYPE_FIRE:

                    break;

		  case SAR_OBJ_TYPE_EXPLOSION:
                    obj_explosion_ptr = obj_ptr->data;
                    if(obj_explosion_ptr == NULL)
                        break;

		    break;

		  case SAR_OBJ_TYPE_HUMAN:
		    obj_human_ptr = obj_ptr->data;
		    if(obj_human_ptr == NULL)
			break;

		    free(obj_human_ptr->mesg_hoist_in);
		    break;

		  case SAR_OBJ_TYPE_PREMODELED:
		    obj_premodeled_ptr = obj_ptr->data;
                    if(obj_premodeled_ptr == NULL)
                        break;

		    break;

		  default:
		    break;
		}

		free(obj_ptr->data);
		obj_ptr->data = NULL;
	    }

	    /* Deallocate standard visual models. */
            vmodel = &obj_ptr->visual_model_dawn;       /* Dawn. */
            DO_UNREF_VISUAL_MODEL
            vmodel = &obj_ptr->visual_model_night;      /* Night. */
            DO_UNREF_VISUAL_MODEL
            vmodel = &obj_ptr->visual_model_dusk;       /* Dusk. */
            DO_UNREF_VISUAL_MODEL
            vmodel = &obj_ptr->visual_model;            /* Day. */
            DO_UNREF_VISUAL_MODEL


	    /* Deallocate contact bounds structure. */
	    if(obj_ptr->contact_bounds != NULL)
	    {
		free(obj_ptr->contact_bounds);
		obj_ptr->contact_bounds = NULL;
	    }

	    /* Deallocate lights. */
	    SARObjDeleteLights(
		scene,
		&obj_ptr->light, &obj_ptr->total_lights
	    );

	    /* Deallocate name. */
	    free(obj_ptr->name);
	    obj_ptr->name = NULL;

	    /* Deallocate object structure itself. */
	    free(obj_ptr);
	    (*ptr)[n] = NULL;


	    /* Object has now been destroyed, now we need to check if
	     * any other objects are referancing this object. If so we
	     * need to reset those referances.
	     */

	    /* Check if any explosion objects are referancing this object. */
	    for(i = 0; i < (*total); i++)
	    {
		obj_ptr2 = (*ptr)[i];
		if(obj_ptr2 == NULL)
		    continue;

		if(obj_ptr2->type == SAR_OBJ_TYPE_EXPLOSION)
		{
                    obj_explosion_ptr = (sar_object_explosion_struct *)
                        obj_ptr2->data;
                    if(obj_explosion_ptr == NULL)
                        continue;

		    if(obj_explosion_ptr->ref_object == n)
			obj_explosion_ptr->ref_object = -1;
		}
	    }

	    /* Unreferance this object from the scene structure. */
	    if(scene != NULL)
	    {
		/* Was this referanced as the player object? */
		if(obj_ptr == scene->player_obj_ptr)
		    scene->player_obj_ptr = NULL;

		if(n == scene->player_obj_num)
		    scene->player_obj_num = -1;

		/* Was this referenced as the camera target? */
		if(n == scene->camera_target)
		    scene->camera_target = -1;

		/* Remove this object from ground list as needed. */
		SARObjRemoveFromGroundList(scene, obj_ptr);

		/* Remove this object from human need rescue list as needed. */
		SARObjRemoveFromHumanRescueList(scene, obj_ptr);
	    }

#undef DO_UNREF_VISUAL_MODEL
#undef DO_STOP_SOUND
#undef DO_FREE_STORAGE
	}
	return;
}

/*
 *	Generates an array of tiles. The call to begin a list and
 *	set up color and texture should already have been made before
 *	calling this function.
 */
void SARObjGenerateTilePlane(
	double min, double max,	/* In meters. */
	double tile_width, double tile_height
)
{
	double x, y, x2, y2;


	/* Begin gl instructions. */
        glBegin(GL_QUADS);

	/* Quadrant 1 is `upper right' and quadrants are numbed
	 * clockwise.
	 */

	/* Quads in quadrant 1. */
	for(y = 0; y < max; y += tile_height)
        {
            for(x = 0; x < max; x += tile_width)
            {
                x2 = x + tile_width;
                y2 = y + tile_height;
                {
                    glNormal3f(0, 1, 0);
                    glTexCoord2f(0, 1 - 0);
                    glVertex3f((GLfloat)x, (GLfloat)0, (GLfloat)-y);
                    glTexCoord2f(1, 1 - 0);
                    glVertex3f((GLfloat)x2, (GLfloat)0, (GLfloat)-y);
                    glTexCoord2f(1, 1 - 1);
                    glVertex3f((GLfloat)x2, (GLfloat)0, (GLfloat)-y2);
                    glTexCoord2f(0, 1 - 1);
                    glVertex3f((GLfloat)x, (GLfloat)0, (GLfloat)-y2);
                }
            }
        }
        /* Quads in quadrant 2. */
        for(y = 0; y > min; y -= tile_height)
        {
            for(x = 0; x < max; x += tile_width)
            {
                x2 = x + tile_width;
                y2 = y - tile_height;
                {
                    glNormal3f(0, 1, 0);
                    glTexCoord2f(0, 1 - 0);
                    glVertex3f((GLfloat)x, 0, (GLfloat)-y2);
                    glTexCoord2f(1, 1 - 0);
                    glVertex3f((GLfloat)x2, 0, (GLfloat)-y2);
                    glTexCoord2f(1, 1 - 1);
                    glVertex3f((GLfloat)x2, 0, (GLfloat)-y);
                    glTexCoord2f(0, 1 - 1);
                    glVertex3f((GLfloat)x, 0, (GLfloat)-y);
               }
           }
       }
       /* Quads in quadrant 3. */
       for(y = 0; y > min; y -= tile_height)
       {
           for(x = 0; x > min; x -= tile_width)
           {
               x2 = x - tile_width;
               y2 = y - tile_height;
               {
                   glNormal3f(0, 1, 0);
                   glTexCoord2f(1, 1 - 1);
                   glVertex3f((GLfloat)x, 0, (GLfloat)-y);
                   glTexCoord2f(0, 1 - 1);
                   glVertex3f((GLfloat)x2, 0, (GLfloat)-y);
                   glTexCoord2f(0, 1 - 0);
                   glVertex3f((GLfloat)x2, 0, (GLfloat)-y2);
                   glTexCoord2f(1, 1 - 0);
                   glVertex3f((GLfloat)x, 0, (GLfloat)-y2);
               }
            }
        }
        /* Quads in quadrant 4. */
        for(y = 0; y < max; y += tile_height)
        {
            for(x = 0; x > min; x -= tile_width)
            {
                x2 = x - tile_width;
                y2 = y + tile_height;
                {
                   glNormal3f(0, 1, 0);
                   glTexCoord2f(1, 1 - 1);
                   glVertex3f((GLfloat)x, 0, (GLfloat)-y2);
                   glTexCoord2f(0, 1 - 1);
                   glVertex3f((GLfloat)x2, 0, (GLfloat)-y2);
                   glTexCoord2f(0, 1 - 0);
                   glVertex3f((GLfloat)x2, 0, (GLfloat)-y);
                   glTexCoord2f(1, 1 - 0);
                   glVertex3f((GLfloat)x, 0, (GLfloat)-y);
                }
            }
        }

	glEnd();

 	return; 
}
