/*
                              Search and Rescue
 */

#ifndef SAR_H
#define SAR_H

#include <stdio.h>

#include <sys/types.h>
#ifndef __MSW__
# include <sys/time.h>
#endif
#include <limits.h>

#include "../include/os.h"
#include "gw.h"
#include "stategl.h"
#include "obj.h"
#include "weather.h"
#include "gctl.h"
#include "menu.h"
#include "sound.h"
#include "mission.h"
#include "texturelistio.h"
#include "sarfps.h"

#include "config.h"



/*
 *	Program name:
 */
#define PROG_NAME	"Search and Rescue"

/*
 *	Program version:
 */
#define PROG_VERSION		"0.7b"
#define PROG_VERSION_MAJOR	0
#define PROG_VERSION_MINOR	7

/*
 *	Program usage message:
 */
#define PROG_USAGE_MESG	"\
Usage: SearchAndRescue [options] [GUI_options]\n\
\n\
    [options] can be any of the following:\n\
\n\
        --rcfile <file>         Load configuration from <file>.\n\
        -f                      Same as --rcfile.\n\
        --control <type>        Specify which controller to use:\n\
                                   keyboard\n\
                                   joystick\n\
        -c                      Same as --control.\n\
	--display <address>     Specifies display address.\n\
	--recorder <address>    Specifies recorder address.\n\
	--nosound		Do not connect to sound server.\n\
	--nomenubg              No menu background images.\n\
        --console_quiet         Do not print routine messages to stdout.\n\
        --runtime_debug         Print runtime detection messages to stdout.\n\
        --internal_debug        Print internal (terse) messages to stdout.\n\
        --help                  Prints (this) help screen and exits.\n\
        --version               Prints version information and exits.\n\
\n\
    Most customizations can be performed in the options menu.\n\
\n\
    Command line options override any options in the configuration file.\n\
\n\
    [GUI_options] can be any options standard to your GUI, consult your\n\
    GUI's manual for available options.\n\
\n"


/*
 *	Program copyright:
 */
#define PROG_COPYRIGHT  "\
Copyright (C) 1999-2001 WolfPack Entertainment.\n\
This program is protected by international copyright laws and treaties,\n\
distribution and/or modification of this software in violation of the\n\
GNU Public License is strictly prohibited. Violators will be prosicuted\n\
to the fullest extent of the law."

/*
 *	Menu system op codes (for button id codes):
 */
#define SAR_OP_MENU_TO_MAIN		0
#define SAR_OP_MENU_EXIT		1

#define SAR_OP_MENU_TO_MISSION		10
#define SAR_OP_MENU_TO_CAMPAIGN		11
#define SAR_OP_MENU_TO_FREE_FLIGHT      12
#define SAR_OP_MENU_TO_OPTIONS		13

#define SAR_OP_MENU_TO_MISSION_BRIEF		20
#define SAR_OP_MENU_TO_MISSION_BEGIN		21
#define SAR_OP_MENU_TO_MISSION_MAP		22

#define SAR_OP_MENU_TO_FREE_FLIGHT_AIRCRAFT	40
#define SAR_OP_MENU_TO_FREE_FLIGHT_BEGIN	41
#define SAR_OP_MENU_TO_FREE_FLIGHT_WEATHER	42

#define SAR_OP_MENU_TO_OPTIONS_SIMULATION	50
#define SAR_OP_MENU_TO_OPTIONS_CONTROLLER	51
#define SAR_OP_MENU_TO_OPTIONS_CONTROLLER_JS_BTN	52
#define SAR_OP_MENU_TO_OPTIONS_GRAPHICS		53
#define SAR_OP_MENU_TO_OPTIONS_SOUNDS		54


/*
 *	Menu names:
 */
#define SAR_MENU_NAME_MAIN			"Main"

#define SAR_MENU_NAME_FREE_FLIGHT		"Free Flight"
#define SAR_MENU_NAME_FREE_FLIGHT_AIRCRAFT	"Free Flight Aircraft"
#define SAR_MENU_NAME_FREE_FLIGHT_WEATHER	"Free Flight Weather"

#define SAR_MENU_NAME_MISSION			"Mission"
#define SAR_MENU_NAME_MISSION_BRIEF		"Mission Brief"
#define SAR_MENU_NAME_MISSION_MAP		"Mission Map"

#define SAR_MENU_NAME_CAMPAIGN			"Campaign"

#define SAR_MENU_NAME_OPTIONS			"Options"
#define SAR_MENU_NAME_OPTIONS_SIMULATION	"Options Simulation"
#define SAR_MENU_NAME_OPTIONS_CONTROLLER	"Options Controler"
#define SAR_MENU_NAME_OPTIONS_CONTROLLER_JS_BTN	"Options Controler JS Buttons"
#define SAR_MENU_NAME_OPTIONS_GRAPHICS		"Options Graphics"
#define SAR_MENU_NAME_OPTIONS_SOUND		"Options Sound"

#define SAR_MENU_NAME_LOADING_SIMULATION	"Loading Simulation"


/*
 *	Options menu object ID codes (should not conflict
 *	with any existing SAR_OP_MENU_* codes). These ID codes
 *	are for various switch and spin objects on the options
 *	menus.
 */
#define SAR_OPT_SELECT_HOIST_CONTACT		510
#define SAR_OPT_SELECT_DAMAGE_RESISTANCE	511
#define SAR_OPT_SELECT_FLIGHT_PHYSICS		512

#define SAR_OPT_SELECT_JS0_AXISES		520
#define SAR_OPT_SELECT_JS1_AXISES		521
#define SAR_OPT_SELECT_JS0_BUTTON_ACTION	522
#define SAR_OPT_SELECT_JS0_BUTTON_NUMBER	523
#define SAR_OPT_SELECT_JS1_BUTTON_ACTION	524
#define SAR_OPT_SELECT_JS1_BUTTON_NUMBER	525

#define SAR_OPT_SELECT_GROUND_TEXTURE		530
#define SAR_OPT_SELECT_OBJECT_TEXTURE		531
#define SAR_OPT_SELECT_CLOUDS			532
#define SAR_OPT_SELECT_ATMOSPHERE               533
#define SAR_OPT_SELECT_MAX_VISIBILITY		534

#define SAR_OPT_SELECT_SOUND_LEVEL		540
#define SAR_OPT_SELECT_MUSIC			541



/*
 *	Options:
 */
typedef struct {

	Boolean menu_backgrounds;

	Boolean	console_quiet,
		internal_debug,
		runtime_debug;

	/* Graphics options. */
	Boolean	textured_ground,
		textured_objects,
		textured_clouds,
		atmosphere;

	/* Maximum drawing visibility, value from 0 to 4.
	 * Computed by; miles = 3 + (option.visibility_max * 3).
	 */
	int visibility_max;

	/* Sound options. */
	Boolean engine_sounds,
		event_sounds,
		voice_sounds,
		music;

	/* HUD color. */
	sar_color_struct hud_color;
	GWFont	*hud_font;

	/* Color of messages. */
	sar_color_struct message_color;
	GWFont	*message_font;


	/* Intervals in milliseconds. */
	time_t	smoke_spawn_int,	/* Smoke puff respawn. */
		explosion_frame_int;	/* Explosion frame inc intvl. */

	time_t	smoke_life_span,	/* Smoke puff life span. */
		crash_explosion_life_span;	/* Explosions by crashes. */

	double rotor_wash_vis_coeff;	/* 0.0 to 1.0. */

	/* Game controller type, one of GCTL_CONTROLLER_*. */
	int gctl_controller_type;

	/* Joystick button mappings. */
	int	js0_btn_rotate,		/* Treat bank to heading axis. */
		js0_btn_air_brakes,
		js0_btn_wheel_brakes,
		js0_btn_zoom_in,
		js0_btn_zoom_out,
		js0_btn_hoist_up,
		js0_btn_hoist_down;

	int	js1_btn_rotate,         /* Treat bank as heading axis. */
                js1_btn_air_brakes,
                js1_btn_wheel_brakes,
                js1_btn_zoom_in,
                js1_btn_zoom_out,
                js1_btn_hoist_up,
                js1_btn_hoist_down;

	/* Game controller options. */
	gctl_flags_t gctl_options;

	/* Simulation difficulty. */
	double hoist_contact_expansion_coeff;	/* Hoist contact radius. */
	double damage_resistance_coeff;
#define FLIGHT_PHYSICS_EASY	0
#define FLIGHT_PHYSICS_MODERATE	1
#define FLIGHT_PHYSICS_REALISTIC	2
	int flight_physics;	/* One of FLIGHT_PHYSICS_*. */


	/* Last run selections. */
	int	last_selected_mission,	/* Matches menu list obj sel pos. */
		last_selected_ffscene,
		last_selected_ffaircraft,
		last_selected_ffweather,
		last_x, last_y, last_width, last_height;

} sar_option_struct;
extern sar_option_struct option;

/*
 *	Directory names:
 */
typedef struct {

	char local_data[PATH_MAX];
	char global_data[PATH_MAX];

} sar_dname_struct;
extern sar_dname_struct dname;

/*
 *	File names:
 */
typedef struct {

	/* Run time preferances. */
	char preferances[PATH_MAX + NAME_MAX];

	/* Basic textures list. */
	char textures[PATH_MAX + NAME_MAX];

	/* Weather preset definations. */
	char weather[PATH_MAX + NAME_MAX];

} sar_fname_struct;
extern sar_fname_struct fname;


/*
 *	Timmers:
 */
typedef struct {

	time_t spawn_smoke;	/* Create smoke puff. */

} sar_next_struct;
extern sar_next_struct next;


/*
 *	Core structure:
 */
typedef struct {

	/* Video, sound, and game controller. */
	gw_display_struct *display;		/* Graphics wrapper. */
	snd_recorder_struct *recorder;		/* Sound server. */
        gctl_struct *gctl;			/* Game controller. */

	/* Current audio mode name the sound server is in. */
	char *audio_mode_name;

	/* Weather data. */
	sar_weather_data_struct *weather_data;

	/* SAR scenery, objects, and mission. */
	sar_scene_struct *scene;		/* SAR scene. */
	sar_mission_struct *mission;		/* SAR mission. */
	sar_object_struct **object;		/* SAR objects. */
	int total_objects;


	/* Menu and shared data for menus. */
	sar_menu_struct **menu;
	int total_menus, cur_menu;

	sar_image_struct	*menu_list_bg_img,
				*menu_button_armed_img,
				*menu_button_unarmed_img,
				*menu_button_highlighted_img,
				*menu_label_bg_img,
				*menu_switch_bg_img,
				*menu_switch_off_img,
				*menu_switch_on_img,
				*menu_spin_label_img,
				*menu_spin_value_img,
				*menu_progress_bg_img,
				*menu_progress_fg_img,

				*menumap_helipad_img,
				*menumap_intercept_img,	/* Interecpt point. */
				*menumap_helicopter_img,
				*menumap_victim_img,
				*menumap_boat_img;

	/* Global list of texture referance names and file names
	 * that are to be loaded to a scene when the scene is loaded.
	 */
	sar_texture_name_struct **texture_list;
	int total_texture_list;

	/* Selected files. */
	char *cur_mission_file;
	char *cur_player_model_file;
	char *cur_scene_file;

	/* Input and display states. */
	Boolean	display_help;

	/* Text prompt input callback function, when this pointer
	 * is not NULL, text is gathered into the text input buffer.
	 * When enter is pressed, the callback function is called
	 * and then set back to NULL.
	 */
	char *text_input_prompt, *text_input_buffer;
	void (*text_input_cb)(void *, const char *);

	/* Frames per second counter and tallying structure. Note
	 * that we have one fps structure per graphics wrapper structure
	 * (which we only have one) so we only have one fps structure.
	 */
	sar_fps_struct fps;

} sar_core_struct;


/*
 *	Client data for menu list object items' client data.
 */
typedef struct {

	sar_core_struct *core_ptr;

	/* For aircraft, mission, and scene listings. */
	char *filename;

	/* For scene start locations. */
	char *name;
	sar_position_struct pos;
	sar_direction_struct dir;

} sar_menu_list_item_data_struct;






extern double	debug_value;

extern int	runlevel;

extern time_t	cur_millitime,
		cur_systime,
		lapsed_millitime;

extern double	time_compensation,
		time_compression;



/* In main.c */
extern void SARHandleSignal(int s);
extern void SARUpdateBackgroundMusic(sar_core_struct *core_ptr);
extern int SARInitGCTL(sar_core_struct *core_ptr);
/* DELETE THIS LATER. */
extern void SARTextInputCBSendMessage(void *client_data, const char *buffer);
extern void SARTextInputCBQuitSimulation(void *client_data, const char *buffer);
extern void SARDrawCB(void *ptr);
extern void SARReshapeCB(void *ptr, int x, int y, int width, int height);
extern void SARKeyBoardCB(void *ptr, int c, Boolean state, long t);
extern void SARPointerCB(void *ptr, int x, int y, int state, int btn_num, long t);
extern void SARVisibilityCB(void *ptr, int v);
extern void SARCloseCB(void *ptr, void *data);
extern void SARResetTimmersCB(sar_core_struct *core_ptr, time_t t_new);

extern sar_core_struct *SARInit(int argc, char **argv);
extern void SARManage(void *ptr);
extern void SARShutdown(sar_core_struct *core_ptr);


/* In mission.c (mission callbacks, utils, and procedures) */
extern void SARMissionPrintStats(
	sar_core_struct *core_ptr,
	sar_scene_struct *scene,
	sar_mission_struct *mission,
	sar_object_struct *obj_ptr
);
extern void SARMissionDestroyNotify(
	sar_core_struct *core_ptr,
	sar_object_struct *obj_ptr
);
extern void SARMissionLandNotify(
	sar_core_struct *core_ptr,
        sar_object_struct *obj_ptr
);
extern int SARMissionManage(sar_core_struct *core_ptr);

/* In missionio.c (mission file io) */
extern void SARMissionDestroy(sar_mission_struct *mission);
extern char *SARMissionLoadDescription(const char *filename);
extern void SARMissionLoadMenuMap(
        sar_core_struct *core_ptr,
        sar_menu_struct *m,
        const char *filename          /* Mission file name. */
);
extern sar_mission_struct *SARMissionLoadFromFile(
	sar_core_struct *core_ptr,
        const char *filename,
        void *client_data,
        int (*progress_func)(void *, long, long)
);

/* In objio.c (object model file io) */
extern char *SARObjFSeekGetOpString(FILE *fp);
extern int SARObjLoadSound(
        snd_recorder_struct *recorder,
        snd_play_struct **snd_play_rtn, char **snd_path_rtn,
        const char *path, int options
);
extern int SARObjLoadLineTexture(
        sar_core_struct *core_ptr, sar_scene_struct *scene,
	const char *line
);
extern int SARObjLoadLineHeightField(
        sar_core_struct *core_ptr, sar_scene_struct *scene,
        const char *line,
	sar_object_struct *obj_ptr
);
extern int SARObjLoadFromFile(
        sar_core_struct *core_struct,
        int n,                  /* Object number. */
        const char *filename
);

/* In objiocb.c (object model file io callbacks) */
extern int SARSceneLoadProgressCB(
	void *ptr,
	long file_pos, long file_size
);

/* In objiopremodeled.c (sar premodeled object io/creation) */
extern int SARObjCreatePremodeled(
	sar_core_struct *core_ptr,
	sar_scene_struct *scene,
	int argc, char **argv
);

/* In prefio.c (options/preferances file io) */
extern int SARLoadPreferances(sar_option_struct *opt, const char *filename);
extern int SARSavePreferances(sar_option_struct *opt, const char *filename);

/* In sardraw.c (primarly simulation drawing) */
extern void SARDraw(sar_core_struct *core_ptr);

/* In sardrawpremodeled.c (drawing of sar premodeled objects) */
extern void SARDrawPremodeled(
	gw_display_struct *display,
	sar_scene_struct *scene,
	sar_object_struct *obj_ptr,
	sar_object_premodeled_struct *obj_premodeled_ptr,
	sar_position_struct *camera_pos,
	sar_direction_struct *camera_dir
);

/* In sarinstall.c (checking and installation of data files) */
extern Boolean SARIsInstalledLocal(
        sar_dname_struct *dn,
        sar_fname_struct *fn
);
extern Boolean SARIsInstalledGlobal(
        sar_dname_struct *dn,
        sar_fname_struct *fn
);
extern int SARDoInstallLocal(
        sar_dname_struct *dn,
        sar_fname_struct *fn,
        sar_option_struct *opt
);

/* In sarkey.c (sar simulation keyboard key event handling) */
extern void SARKey(
	sar_core_struct *core_ptr,
	int c,		/* GW Key value or character. */
	Boolean state,	/* Pressed (True) or released (False). */
	long t		/* Time stamp. */
);

/* In sarmenubuild.c (creates all menus) */
extern int SARBuildMenus(sar_core_struct *core_ptr);

/* In sarmenucb.c (menu object/widget callbacks) */
extern void SARMenuButtonCBSelectMenu(
        sar_core_struct *core_ptr,
        const char *name
);
extern sar_menu_list_struct *SARMenuGetList(
	sar_menu_struct *menu_ptr, int skip, int *list_num
);
extern void SARMenuButtonCB(
	void *object, void *client_data,
	int id_code
);
extern void SARMenuListSelectCB(
        void *object, void *client_data,
        int item_num, void *item_ptr
);
extern void SARMenuSpinSelectCB(
        void *object, void *client_data,
        int id_code, char *value
);

/* In saroptions.c (sar options/preferances procedures and callbacks) */
extern void SARApplyOptionsFromMenus(sar_core_struct *core_ptr);
extern void SARFetchOptionsToMenus(sar_core_struct *core_ptr);

extern void SARMenuOptionsSwitchCB(
        void *object, void *client_data,
        int id_code, Boolean state
);
extern void SARMenuOptionsSpinCB(
        void *object, void *client_data,
        int id_code, char *value
);

/* In sarsimend.c (sar simulation end tabulation and procedures) */
extern void SAREndSimulation(sar_core_struct *core_ptr);

/* In sarutils.c (sar specific utility functions) */
extern time_t SARGetCurMilliTime(void);
extern int SARRandom(void);

extern Boolean SARGetGLVersion(int *major, int *minor, int *release);
extern char *SARGetGLVendorName(void);
extern char *SARGetGLRendererName(void);
extern char **SARGelGLExtensionNames(int *strc);

extern int SARIsMenuAllocated(sar_core_struct *core_ptr, int n);
extern int SARMatchMenuByName(
        sar_core_struct *core_ptr,
        const char *name
);
extern sar_menu_struct *SARGetCurrentMenuPtr(sar_core_struct *core_ptr);
extern void SARDeleteListItemData(
        sar_menu_list_item_struct *item_ptr
);


/* In sceneio.c (scenery file io) */
extern void SARSceneDestroy(    
        sar_core_struct *core_ptr,
        sar_scene_struct *scene,  
        sar_object_struct ***ptr,
        int *total
);
extern void SARSceneLoadLocationsToList(
        sar_core_struct *core_ptr,
        sar_menu_struct *menu_ptr,
        sar_menu_list_struct *list_ptr,
        int list_num,
        const char *filename
);
extern int SARSceneAddPlayerObject(
        sar_core_struct *core_ptr,
        sar_scene_struct *scene,
        const char *model_file,
        sar_position_struct *pos,
        sar_direction_struct *dir
);
extern int SARSceneLoadFromFile(
        sar_core_struct *core_struct,   
        sar_scene_struct *scene,
        const char *filename,
        const char *weather_preset_name,
        void *client_data,
        int (*progress_func)(void *, long, long)
);

/* In scenesound.c (scene sound reloading/updating) */
extern void SARSceneSoundUpdate(
        sar_core_struct *core_ptr,
        Boolean engine_sounds,
        Boolean event_sounds,
        Boolean voice_sounds,
        Boolean music
);



#endif	/* SAR_H */
