/*
                            Weather Preset Data
 */

#ifndef WEATHER_H
#define WEATHER_H

#include <sys/types.h>
#include "obj.h"

/*
 *	Weather data structure:
 */
typedef struct {

	char *name;

	/* Sky and horizon gradient colors. */
	sar_color_struct	sky_nominal_color,
				sky_brighten_color,
				sky_darken_color;

	/* Close celestial object colors. */
	sar_color_struct	star_color,
				moon_color;

	double	atmosphere_dist_coeff,
		atmosphere_density_coeff;

	sar_cloud_layer_struct **cloud_layer;
	int total_cloud_layers;

} sar_weather_data_entry_struct;


/*
 *	Contains a list of preset weather data structures.
 */
typedef struct {

	void *core_ptr;		/* Pointer back to core. */

	sar_weather_data_entry_struct **preset;
	int total_presets;

} sar_weather_data_struct;




extern sar_weather_data_struct *SARWeatherInit(void *core_ptr);
extern void SARWeatherShutdown(sar_weather_data_struct *w);

extern void SARWeatherEntryDelete(
	sar_weather_data_struct *w, sar_weather_data_entry_struct *entry
);

extern void SARWeatherSetScenePreset(
        sar_weather_data_struct *w,
        sar_scene_struct *scene,
        const char *name
);

extern int SARWeatherLoadFromFile(
	sar_weather_data_struct *w,
	const char *filename
);


#endif	/* WEATHER_H */
